@@ -1,21 +1,30 @@
 #!/usr/bin/env python
-"""Convenience script wrapper to start and stop AiiDAlab via docker-compose.
+"""Tool to launch and manage AiiDAlab instances with docker.
 
 Authors:
     * Carl Simon Adorf <simon.adorf@epfl.ch>
 """
+import getpass
 import logging
+import socket
 from dataclasses import dataclass, field
 from pathlib import Path
+from textwrap import wrap
 
 import click
 import docker
 from tabulate import tabulate
 
-from .core import APPLICATION_ID, LOGGER, AiidaLabInstance, Config, Profile, Timeout
-from .util import get_docker_client, spinner
+from .core import APPLICATION_ID, LOGGER, AiidaLabInstance, Config, Profile
+from .util import get_docker_client, spinner, webbrowser_available
 from .version import __version__
 
+MSG_MOUNT_POINT_CONFLICT = """Warning: There is at least one other running
+instance that has the same home mount point ('{home_mount}') as the instance
+you are currently trying to start. Are you sure you want to continue? This may
+lead to data corruption."""
+
+
 MSG_STARTUP = """Open the following URL to access AiiDAlab:
 
   {url}
@@ -23,6 +32,24 @@ MSG_STARTUP = """Open the following URL to access AiiDAlab:
 Home mounted: {home_mount} -> /home/{system_user}"""
 
 
+MSG_STARTUP_SSH = """
+Unable to detect a web browser which indicates that you might be running
+AiiDAlab on a remote machine. If this is the case, consider to create an SSH
+tunnel to access AiiDAlab on your local computer. For this, run a command
+similar to
+
+  ssh {user}@{hostname} -NfL {port}:localhost:{port}
+
+on your local computer, then open AiiDAlab on your local computer at
+
+  {url}
+
+See https://github.com/aiidalab/aiidalab-launch/blob/main/ssh-forward.md for
+more detailed instructions on SSH port forwarding.
+
+Home mounted: {home_mount} -> /home/{system_user}"""
+
+
 LOGGING_LEVELS = {
     0: logging.ERROR,
     1: logging.WARN,
@@ -53,7 +80,7 @@ pass_app_state = click.make_pass_decorator(ApplicationState, ensure=True)
 
 
 def with_profile(cmd):
-    def callback(ctx, param, value):
+    def callback(ctx, param, value):  # noqa: U100
         app_state = ctx.ensure_object(ApplicationState)
         name = value or app_state.config.default_profile
         LOGGER.info(f"Using profile: {name}")
@@ -73,16 +100,15 @@ def with_profile(cmd):
 )
 def cli(verbose):
     # Use the verbosity count to determine the logging level...
+    logging.basicConfig(
+        level=LOGGING_LEVELS[verbose] if verbose in LOGGING_LEVELS else logging.DEBUG
+    )
     if verbose > 0:
-        logging.basicConfig(
-            level=LOGGING_LEVELS[verbose]
-            if verbose in LOGGING_LEVELS
-            else logging.DEBUG
-        )
         click.secho(
             f"Verbose logging is enabled. "
             f"(LEVEL={logging.getLogger().getEffectiveLevel()})",
             fg="yellow",
+            err=True,
         )
 
     LOGGER.info(f"Configuration file path: {_application_config_path()}")
@@ -152,17 +178,21 @@ def add_profile(ctx, app_state, profile):
 
 @profiles.command("remove")
 @click.argument("profile")
+@click.option("--yes", is_flag=True, help="Do not ask for confirmation.")
 @pass_app_state
-def remove_profile(app_state, profile):
+def remove_profile(app_state, profile, yes):
     """Remove a AiiDAlab profile from the configuration."""
     try:
         profile = app_state.config.get_profile(profile)
     except ValueError:
         raise click.ClickException(f"Profile with name '{profile}' does not exist.")
     else:
-        app_state.config.profiles.remove(profile)
-        app_state.config.save(_application_config_path())
-        click.echo(f"Removed profile with name '{profile.name}'.")
+        if yes or click.confirm(
+            f"Are you sure you want to remove profile '{profile.name}'?"
+        ):
+            app_state.config.profiles.remove(profile)
+            app_state.config.save(_application_config_path())
+            click.echo(f"Removed profile with name '{profile.name}'.")
 
 
 @profiles.command("edit")
@@ -197,6 +227,24 @@ def set_default_profile(app_state, profile):
         click.echo(f"Set default profile to '{profile}'.")
 
 
+def _find_mount_point_conflict(client, profile, other_profiles):
+    """Find running instances with the same home mount point.
+
+    To protect users from inadvertently starting a second profile with the same
+    home mount point. Running two containers with the same home mount point has
+    potential for data corruption.
+    """
+    for other_profile in other_profiles:
+        if (
+            other_profile != profile
+            and Path(other_profile.home_mount).resolve()
+            == Path(profile.home_mount).resolve()
+            and AiidaLabInstance(client=client, profile=other_profile).status()
+            is not AiidaLabInstance.AiidaLabInstanceStatus.DOWN
+        ):
+            yield other_profile
+
+
 @cli.command()
 @click.option(
     "--restart",
@@ -226,45 +274,99 @@ def set_default_profile(app_state, profile):
         "(This is disabled by default if the wait time is set to zero.)"
     ),
 )
+@click.option(
+    "--show-ssh-port-forwarding-help",
+    "show_ssh_help",
+    is_flag=True,
+    help="Show guidance on SSH port forwarding.",
+)
+@click.option(
+    "-f", "--force", is_flag=True, help="Ignore any warnings and start anyways."
+)
 @pass_app_state
 @with_profile
-def start(app_state, profile, restart, wait, pull, no_browser):
+def start(app_state, profile, restart, wait, pull, no_browser, show_ssh_help, force):
     """Start an AiiDAlab instance on this host."""
 
     instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
-    try:
-        if instance.container() is None:
-            if pull:
-                with spinner(
-                    f"Downloading image '{instance.profile.image}' (this may take a while)..."
-                ):
-                    instance.pull()
-            else:
-                raise click.ClickException(
-                    f"Unable to find container '{instance.profile.container_name()}'. "
-                    "Use '--pull' to pull the image prior to start."
+
+    # Check for potential mount point conflicts.
+    if not force:
+        with spinner("Check for potential conflicts...", delay=0.1):
+            conflict = any(
+                _find_mount_point_conflict(
+                    app_state.docker_client,
+                    profile,
+                    app_state.config.profiles,
                 )
+            )
+        if conflict:
+            msg_warn = MSG_MOUNT_POINT_CONFLICT.format(home_mount=profile.home_mount)
+            click.confirm(
+                click.style("\n".join(wrap(msg_warn)), fg="yellow"),
+                abort=True,
+            )
+
+    # Obtain image (either via pull or local).
+    if pull:
+        try:
+            msg = (
+                f"Downloading image '{instance.profile.image}', this may take a while..."
+                if instance.image is None
+                else f"Downloading latest version of '{instance.profile.image}'..."
+            )
+            with spinner(msg):
+                instance.pull()
+        except RuntimeError as error:
+            raise click.ClickException(str(error))
+    elif instance.image is None:
+        raise click.ClickException(
+            f"Unable to find image '{profile.image}'. "
+            "Try to use '--pull' to pull the image prior to start."
+        )
+
+    # Check if image has changed.
+    assert instance.image is not None
+    recreate = instance.container and instance.container.image.id != instance.image.id
+
+    try:
 
         InstanceStatus = instance.AiidaLabInstanceStatus  # local alias for brevity
 
         status = instance.status()
-        if status is InstanceStatus.DOWN:
+        if status in (
+            InstanceStatus.DOWN,
+            InstanceStatus.CREATED,
+            InstanceStatus.EXITED,
+        ):
+            if recreate:
+                with spinner("Recreating container..."):
+                    instance.remove()
+                    instance.create()
             with spinner("Starting container..."):
                 instance.start()
-        elif status is InstanceStatus.CREATED:
-            with spinner("Starting previously created container..."):
-                instance.restart()
         elif status is InstanceStatus.UP and restart:
             with spinner("Restarting container..."):
-                instance.restart()
+                if recreate:
+                    instance.stop()
+                    instance.remove()
+                instance.start()
         elif status is InstanceStatus.UP and not restart:
-            click.echo(
-                "Container was already running, use --restart to restart it.", err=True
-            )
+            if recreate:
+                click.secho(
+                    "Container is already running, however the image has changed. "
+                    "A restart with --restart is recommended.",
+                    fg="yellow",
+                )
+            else:
+                click.echo(
+                    "Container was already running, use --restart to restart it.",
+                    err=True,
+                )
         elif status is InstanceStatus.STARTING:
             click.echo("Container is already starting up...", err=True)
 
-    except Timeout:
+    except TimeoutError:
         raise click.ClickException(
             f"AiiDAlab instance did not start up within the provided wait period ({wait})."
         )
@@ -284,22 +386,38 @@ def start(app_state, profile, restart, wait, pull, no_browser):
         raise click.ClickException(f"Unknown error occurred: {error}")
     else:
         if wait:
-            with spinner("Waiting for AiiDAlab services to start..."):
-                instance.wait_for_services(timeout=wait)
+            try:
+                with spinner("Waiting for AiiDAlab instance to get ready..."):
+                    instance.wait_for_services(timeout=wait)
+            except RuntimeError:
+                raise click.ClickException(
+                    "The AiiDAlab instance failed to start. Consider to inspect "
+                    "the container output logs by increasing the output "
+                    "verbosity with 'aiidalab-launch -v start'."
+                )
             url = instance.url()
+            msg_startup = (
+                MSG_STARTUP_SSH
+                if (show_ssh_help or not webbrowser_available())
+                else MSG_STARTUP
+            )
             click.secho(
-                MSG_STARTUP.format(
+                msg_startup.format(
                     url=instance.url(),
                     home_mount=instance.profile.home_mount,
                     system_user=instance.profile.system_user,
-                ),
+                    user=getpass.getuser(),
+                    port=instance.profile.port,
+                    hostname=socket.getfqdn(),
+                ).lstrip(),
                 fg="green",
             )
-            if not no_browser:
+            if not no_browser and webbrowser_available():
                 if click.confirm(
                     "Do you want to open AiiDAlab in the browser now?", default=True
                 ):
                     click.launch(url)
+
         else:
             click.secho(
                 "Use 'aiidalab-launch status' to check the AiiDAlab instance "
@@ -392,20 +510,32 @@ def status(app_state):
 @click.argument("cmd", nargs=-1)
 @click.option("-p", "--privileged", is_flag=True)
 @click.option("--forward-exit-code", is_flag=True)
+@click.option(
+    "--wait/--no-wait",
+    help="Wait on AiiDAlab services to get ready before sending the command.",
+    show_default=True,
+    default=True,
+)
 @click.pass_context
 @with_profile
-def exec(ctx, profile, cmd, privileged, forward_exit_code):
+def exec(ctx, profile, cmd, privileged, forward_exit_code, wait):
     """Directly execute a command on a AiiDAlab instance.
 
     For example, to get a list of all installed aiidalab applications, run:
 
-        aiidalab-launch exec aiidalab list
+        aiidalab-launch exec -- aiidalab list
 
     """
     app_state = ctx.ensure_object(ApplicationState)
     instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
-
-    exec_id = instance.exec_create(" ".join(cmd), privileged=privileged)
+    try:
+        if wait:
+            with spinner("Waiting for AiiDAlab to get ready...", delay=0.5):
+                instance.wait_for_services(timeout=60)
+        with spinner("Send command to container...", delay=1.0):
+            exec_id = instance.exec_create(" ".join(cmd), privileged=privileged)
+    except RuntimeError:
+        raise click.ClickException("AiiDAlab instance is not available. Is it running?")
 
     output = app_state.docker_client.api.exec_start(exec_id, stream=True)
     for chunk in output:
